import { ref, reactive, watch, h, defineComponent, onMounted } from 'vue';
import { v4 } from 'uuid';
import * as _ from 'lodash';
import Draggable from 'vuedraggable';

const useIndex = function useIndex({ load, filters = {}, sortBy = [], }) {
    const search = ref('');
    let index = reactive({
        isBusy: false,
        items: [],
        meta: {},
        search,
        filters,
        sortBy,
        async load() {
            this.isBusy = true;
            return load(this.__getParams())
                .then(response => {
                this.items = response.data.data;
                this.meta = response.data.meta;
                return new Promise(() => response);
            })
                .finally(() => (this.isBusy = false));
        },
        reloadOnChange(value) {
            watch(value, () => this.load());
        },
        setFilter(type, value) {
            if (!this.filters[type]) {
                this.filters[type] = {};
            }
            this.filters[type].value = value;
            this.load();
        },
        removeFilter(type) {
            const value = this.filters[type].value;
            if (Array.isArray(value)) {
                this.filters[type] = [];
            }
            else if (typeof value === 'object') {
                this.filters[type] = {};
            }
            else {
                this.filters[type] = undefined;
            }
            this.load();
        },
        setPage(newPage) {
            if (this.newPage < 1 || this.newPage > this.lastPage) {
                return;
            }
            this.currentPage = newPage;
            this.load();
        },
        setNextPage() {
            this.setPage(this.currentPage + 1);
        },
        setPrevPage() {
            this.setPage(this.currentPage - 1);
        },
        setLastPage() {
            this.setPage(this.getLastPage());
        },
        updateSearch(e) {
            search.value =
                e instanceof Event
                    ? e.target.value
                    : e;
            this.setPage(1);
        },
        isSortedBy(key, direction = null) {
            if (!this.sortBy.includes(key))
                return false;
            if (!this.sortBy[key])
                return false;
            if (!direction)
                return true;
            return this.sortBy[key] == direction;
        },
        addSortBy(key, direction = 'asc') {
            this.sortBy.push({ key, direction });
            this.load();
        },
        removeSortBy(key) {
            for (let i = 0; i < this.sortBy.length; i++) {
                if (this.sortBy[i].key == key)
                    delete this.sortBy[i];
            }
            this.load();
        },
        __getParams() {
            let params = {
                page: this.currentPage,
                search: this.search,
                sortBy: [],
            };
            for (let i = 0; i < this.sortBy; i++) {
                params.sortBy.push(`${this.sortBy[i].key}.${this.sortBy[i].direction}`);
            }
            for (let key in this.filters) {
                if (!this.filters[key]) {
                    continue;
                }
                params[`filter.${key}`] =
                    'transform' in this.filters[key]
                        ? this.filters[key].transform(this.filters[key].value)
                        : this.filters[key].value;
            }
            for (let key in params) {
                if (!params[key])
                    delete params[key];
            }
            return params;
        },
    });
    index.reloadOnChange(search);
    return index;
};

const clone = value => {
    return JSON.parse(JSON.stringify(value));
};
const useOriginal = value => {
    const original = reactive({
        raw: clone(value),
        stringified: JSON.stringify(value),
        update(value) {
            this.raw = clone(value);
            this.stringified = JSON.stringify(value);
        },
        matches(value) {
            return JSON.stringify(value) == this.stringified;
        },
    });
    return original;
};

const useForm = function ({ data, submit, load = async (id) => { }, transform = data => data, onDirty = () => { }, }) {
    let form = reactive({
        ...data,
        errors: {},
        isDirty: false,
        isSubmitting: false,
        isLoading: false,
        original: useOriginal(data),
        __id: undefined,
        load(id) {
            this.isLoading = true;
            this.__id = id;
            return load(id)
                .then(response => {
                this.setData(response.data.data);
                this.isLoading = false;
                return response;
            })
                .catch(e => {
                this.isLoading = false;
                throw e;
            });
        },
        data() {
            return Object.keys(data).reduce((carry, key) => {
                carry[key] = this[key];
                return carry;
            }, {});
        },
        setData(newData) {
            Object.keys(data).forEach(key => (this[key] = newData[key]));
            this.original.update(this.data());
            this.isDirty = false;
        },
        submit(e) {
            if (e instanceof Event) {
                e.preventDefault();
            }
            this.isSubmitting = true;
            const data = transform(this.data());
            return submit(data, this.__id)
                .then(response => {
                this.errors = {};
                this.original.update(this.data());
                this.isSubmitting = false;
                return response;
            })
                .catch(e => {
                let data = e.response.data;
                if ('errors' in data) {
                    this.errors = data.errors;
                }
                this.isSubmitting = false;
                throw e;
            });
        },
    });
    onDirty.bind(form);
    watch(form, () => {
        form.isDirty = !form.original.matches(form.data());
        if (form.isDirty) {
            onDirty(form);
        }
    }, { immediate: true, deep: true });
    return form;
};

function parseOrderRecursive(list) {
    let order = [];
    for (let i = 0; i < list.items.length; i++) {
        order.push({
            id: list.items[i].value.id,
            children: parseOrderRecursive(list.items[i].children),
        });
    }
    return order;
}
const useTree = ({ items = [], load = async (params) => { } }) => {
    const list = reactive({
        items: [],
        isBusyLoading: false,
        __changeHandlers: [],
        onOrderChange(handler) {
            this.__changeHandlers.push(handler);
        },
        load(params) {
            if (!load)
                throw new Error('Missing load function for tree.');
            this.isBusyLoading = true;
            return load(params)
                .then(response => {
                this.setItems(response.data.data);
                this.isBusyLoading = false;
                return new Promise(() => response);
            })
                .catch(e => {
                this.isBusyLoading = false;
                throw e;
            });
        },
        push(item, children = []) {
            this.items.push({
                children: useTree({ items: children }),
                value: item,
            });
        },
        pop() {
            return this.items.pop();
        },
        setItems(list) {
            let items = [];
            for (let i = 0; i < list.length; i++) {
                items.push({
                    children: useTree({ items: list[i].children }),
                    uuid: v4(),
                    value: list[i].value,
                });
            }
            this.items = items;
        },
        getOrder() {
            return parseOrderRecursive(this);
        },
    });
    list.setItems(items);
    list.updateOnChange = items => {
        watch(items, () => {
            list.setItems(typeof items === 'function' ? items() : items);
        }, { immediate: true, deep: true });
    };
    const originalOrder = useOriginal(list.getOrder());
    watch(() => list, () => {
        const order = list.getOrder();
        if (originalOrder.matches(order))
            return;
        originalOrder.update(order);
        if (list.isBusyLoading)
            return;
        for (let i = 0; i < list.__changeHandlers.length; i++) {
            list.__changeHandlers[i](order);
        }
    }, { immediate: true, deep: true });
    return list;
};

const useState = (value, { load }) => {
    return reactive({
        value,
        isLoading: false,
        isLoaded: false,
        load() {
            this.isLoading = true;
            return load
                .apply(null, arguments)
                .then(response => {
                this.value = response.data.data;
                this.isLoading = false;
                this.isLoaded = true;
                return this.value;
            })
                .catch(e => {
                this.isLoading = false;
                throw e;
            });
        },
    });
};

const useSaveQueue = function () {
    const saveQueue = reactive({
        jobs: {},
        isDirty: false,
        add(key, job, callback = () => null) {
            this.jobs[key] = { job, callback };
        },
        remove(key) {
            delete this.jobs[key];
        },
        async save() {
            let promises = [];
            for (let key in this.jobs) {
                let func = async () => {
                    await this.jobs[key].job();
                    this.jobs[key].callback();
                };
                promises.push(func());
            }
            this.jobs = {};
            await Promise.all(promises.map(p => p.catch(e => e)));
        },
    });
    watch(saveQueue, () => {
        saveQueue.isDirty = Object.keys(saveQueue.jobs).length > 0;
    }, {
        // force eager callback execution
        immediate: true
    });
    return saveQueue;
};

function getValue(values, locale) {
    return (values.value)[locale.value] || null;
}
const translatable = (locale, values, onChange) => {
    const vals = ref(values || {});
    const val = ref(getValue(vals, locale));
    watch(vals, (v) => onChange(v), { deep: true });
    watch(locale, () => val.value = getValue(vals, locale), { deep: true });
    watch(val, (v) => vals.value[locale.value] = v);
    return val;
};

const Input = function ({ modelModifiers = {} }, { attrs = {}, emit }) {
    let onInput = ({ target }) => {
        emit('update:modelValue', target.value);
    };
    if (modelModifiers.debounce) {
        onInput = _.debounce(onInput, 200);
    }
    return h(`input`, {
        ...attrs,
        value: attrs.modelValue,
        onInput,
    });
};

const template$1 = `
    <Draggable
        v-if="value"
        v-model="value"
        v-bind="dragOptions"
        :group="group"
        itemKey="uuid"
        :component-data="{
            name: !drag ? 'flip-list' : null,
            wrap: true,
        }"
        @start="drag = true"
        @end="drag = false"
        handle=".handle"
    >
        <template #item="{ element }">
            <div>
                <component
                    :is="sections[getSectionKey(element.component)]"
                    :modelValue="element.value"
                    @update:modelValue="(e) => updateElement(element, e)"
                    @destroy="(e) => removeElement(element)"
                />
            </div>
        </template>
    </Draggable>
`;
const Sections = defineComponent({
    template: template$1,
    components: { Draggable },
    emits: ['update:modelValue'],
    props: {
        modelValue: {
            type: Array,
            required: true,
        },
        sections: {
            type: Object,
            required: true,
        },
        group: {
            type: String,
            default: 'sections',
        },
        dragOptions: {
            type: Object,
            default: () => ({
                animation: 200,
                ghostClass: 'ghost',
            }),
        },
    },
    setup(props, { emit }) {
        const drag = ref(false);
        const updateElement = (element, value) => {
            element.value = value;
        };
        const getSectionKey = (section) => {
            const json = JSON.stringify(section);
            for (let key in props.sections) {
                if (JSON.stringify(props.sections[key]) == json) {
                    return key;
                }
            }
            return '';
        };
        const removeElement = (element) => {
            value.value = value.value.filter((section) => {
                return section.uuid != element.uuid;
            });
        };
        const parseValue = (value) => {
            let b = [];
            for (let i in value) {
                b.push({
                    uuid: v4(),
                    key: value[i].type,
                    value: value[i].value,
                    component: props.sections[value[i].type],
                });
            }
            return b;
        };
        const defaultValue = (value, i) => {
            let defaultValue = value[i].component.props.modelValue.default();
            if (Object.keys(defaultValue).length < 1) {
                console.error('No default keys set in Section');
            }
            return defaultValue;
        };
        const transformValue = (value) => {
            let b = [];
            for (let i in value) {
                b.push({
                    type: getSectionKey(value[i].component),
                    value: value[i].value || defaultValue(value, i),
                });
            }
            return b;
        };
        const value = ref(parseValue(props.modelValue));
        watch(() => value, () => emit('update:modelValue', transformValue(value.value)), { deep: true });
        return { drag, updateElement, removeElement, getSectionKey, value };
    }
});

const template = `
    <Draggable
        v-model="drawers"
        :group="{
            name: group,
            pull: 'clone',
            put: false,
        }"
        :sort="false"
        :clone="cloneSection"
        item-key="id"
    >
        <template #item="{ element }">
            <div>
                <component :is="element" />
            </div>
        </template>
    </Draggable>
`;
const Cabinet = defineComponent({
    template,
    components: { Draggable },
    props: { group: { type: String, default: 'sections' } },
    setup({}, { slots }) {
        const drawers = ref();
        onMounted(() => {
            if (slots.default) {
                drawers.value = slots.default();
            }
        });
        const cloneSection = (el) => ({
            uuid: v4(),
            component: el.props.draws,
        });
        return { cloneSection, drawers };
    }
});

export { Cabinet, Input, Sections, translatable, useForm, useIndex, useOriginal, useSaveQueue, useState, useTree };
